/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : SCPCrossoverUniform.cc                                   */
/*                                                                       */
/*       Date : March 5, 1998                                            */
/* Updated at : March 5, 1998                                            */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

#include <Chromosomes.h>
#include <ChromosomesBag.h>
#include <WeightedRandomPickerSet.h>
#include <tSet.h>

#include "SCPCrossoverUniform.h"
#include "SCPChromosomes.h"
#include "SetCoverProblem.h"
#include "SetCoverSolution.h"

/*************************************************************************/
/*************************************************************************/

SCPCrossoverUniform::SCPCrossoverUniform() { debugOn(); }

/*************************************************************************/
/*************************************************************************/

SCPCrossoverUniform::~SCPCrossoverUniform()
{  cout << " de allocate SCPCrossoverUniform " << endl; }

/*************************************************************************/
/*************************************************************************/

ChromosomesBag SCPCrossoverUniform::apply(Chromosomes* par1, Chromosomes * par2)
{   
    if(debug) cout << " I am in SCPCrossoverUniform::apply() " << endl;
    SCPChromosomes *parent1 = (SCPChromosomes *) par1;
    SCPChromosomes *parent2 = (SCPChromosomes *) par2;
    SetCoverProblem *problem = parent1->getProblem();
    int numOfRows = problem->getNumOfRows();
    int numOfCols = problem->getNumOfCols();
    int i,j; 
    float r;

    if(debug) cout << " 1. I am in SCPCrossoverUniform::apply() " << endl;

    // define offsprings
    SCPChromosomes *offspring1 = new SCPChromosomes(parent1->getProblem());
    SCPChromosomes *offspring2 = new SCPChromosomes(parent1->getProblem());

    if(debug) cout << " 2. I am in SCPCrossoverUniform::apply() " << endl;

    // Build binary representation
    tSet<int> pa1s = parent1->getColumns();
    tSet<int> pa2s = parent2->getColumns();
    tSet<int> p1_and_p2 = pa1s * pa2s;
    tSet<int> p1_p2 = pa1s - pa2s;
    tSet<int> p2_p1 = pa2s - pa1s;

    // No matter what the common bits are gonna be in the offsprings
    for(i=0;i<p1_and_p2.getSize();i++) 
	if( p1_and_p2.getElementAtSafe(i,j)) 
	{ offspring1->addColumn(j);
	  offspring2->addColumn(j);
        }

    for(i=0;i<p1_p2.getSize();i++) 
	if( p1_p2.getElementAtSafe(i,j)) 
	{
           r = 0.01 * (rand()%101);
           if( r>=0.5 ) offspring1->addColumn(j);
	   else offspring2->addColumn(j);

        }// end of if

    for(i=0;i<p2_p1.getSize();i++) 
	if( p2_p1.getElementAtSafe(i,j)) 
	{
           r = 0.01 * (rand()%101);
           if( r>=0.5 ) offspring2->addColumn(j);
	   else offspring1->addColumn(j);

        }// end of if

    repair(offspring1);
    repair(offspring2);

    /*if(debug) 
    { 
       cout << " parent1   : " << *parent1 << endl;
       cout << " parent2   : " << *parent2 << endl;
       cout << " offspring1 : " << *offspring1 << endl;
       cout << " offspring2 : " << *offspring2 << endl;
    }// end of if
    */

    offspring1->reduce();
    offspring2->reduce();

    /*if(debug) 
    {
       ofstream ou1("/tmp/uniform.out",ios::app);
       ou1  << " parent1   : " << *parent1 << endl;
       ou1  << " parent2   : " << *parent2 << endl;
       ou1  << " offspring1 : " << *offspring1 << endl;
       ou1  << " offspring2 : " << *offspring2 << endl;
       ou1.close();
    }// end of if
    */

    ChromosomesBag ret(2);
    ret.addChromosomes(offspring1);
    ret.addChromosomes(offspring2);

    return ret;

}// end of apply()

/*************************************************************************/
/**
    Returns number of offsprings
**/
/*************************************************************************/

SCPCrossoverUniform::apply(Chromosomes* par1, Chromosomes * par2,
			   Chromosomes** offsprings)
{   
    if(debug) cout << " I am in SCPCrossoverUniform::apply() " << endl;
    SCPChromosomes *parent1 = (SCPChromosomes *) par1;
    SCPChromosomes *parent2 = (SCPChromosomes *) par2;
    SetCoverProblem *problem = parent1->getProblem();
    int numOfRows = problem->getNumOfRows();
    int numOfCols = problem->getNumOfCols();
    int i,j; 
    float r;

    if(debug) cout << " 1. I am in SCPCrossoverUniform::apply() " << endl;

    // define offsprings
    SCPChromosomes *offspring1 = (SCPChromosomes*)offsprings[0];
    SCPChromosomes *offspring2 = (SCPChromosomes*)offsprings[1];

    if(debug) cout << " 2. I am in SCPCrossoverUniform::apply() " << endl;

    // Build binary representation
    tSet<int> pa1s = parent1->getColumns();
    tSet<int> pa2s = parent2->getColumns();
    tSet<int> p1_and_p2 = pa1s * pa2s;
    tSet<int> p1_p2 = pa1s - pa2s;
    tSet<int> p2_p1 = pa2s - pa1s;

    // No matter what the common bits are gonna be in the offsprings
    for(i=0;i<p1_and_p2.getSize();i++) 
	if( p1_and_p2.getElementAtSafe(i,j)) 
	{ offspring1->addColumn(j);
	  offspring2->addColumn(j);
        }

    for(i=0;i<p1_p2.getSize();i++) 
	if( p1_p2.getElementAtSafe(i,j)) 
	{
           r = 0.01 * (rand()%101);
           if( r>=0.5 ) offspring1->addColumn(j);
	   else offspring2->addColumn(j);

        }// end of if

    for(i=0;i<p2_p1.getSize();i++) 
	if( p2_p1.getElementAtSafe(i,j)) 
	{
           r = 0.01 * (rand()%101);
           if( r>=0.5 ) offspring2->addColumn(j);
	   else offspring1->addColumn(j);

        }// end of if

    repair(offspring1);
    repair(offspring2);

    /*if(debug) 
    { 
       cout << " parent1   : " << *parent1 << endl;
       cout << " parent2   : " << *parent2 << endl;
       cout << " offspring1 : " << *offspring1 << endl;
       cout << " offspring2 : " << *offspring2 << endl;
    }// end of if
    */

    offspring1->reduce();
    offspring2->reduce();

    /*if(debug) 
    {
       ofstream ou1("/tmp/uniform.out",ios::app);
       ou1  << " parent1   : " << *parent1 << endl;
       ou1  << " parent2   : " << *parent2 << endl;
       ou1  << " offspring1 : " << *offspring1 << endl;
       ou1  << " offspring2 : " << *offspring2 << endl;
       ou1.close();
    }// end of if
    */

    return 2;

}// end of apply()

/*************************************************************************/
/*************************************************************************/

ChromosomesBag SCPCrossoverUniform::applyOLD(Chromosomes* par1, Chromosomes * par2)
{   
    if(debug) cout << " I am in SCPCrossoverUniform::apply() " << endl;
    SCPChromosomes *parent1 = (SCPChromosomes *) par1;
    SCPChromosomes *parent2 = (SCPChromosomes *) par2;
    SetCoverProblem *problem = parent1->getProblem();
    int numOfRows = problem->getNumOfRows();
    int numOfCols = problem->getNumOfCols();
    int i,j; 
    float r;

    Column cpr1(numOfCols);
    Column cpr2(numOfCols);
    Column coff1(numOfCols);
    Column coff2(numOfCols);

    if(debug) cout << " 1. I am in SCPCrossoverUniform::apply() " << endl;

    SCPChromosomes *offspring1 = new SCPChromosomes(parent1->getProblem());
    SCPChromosomes *offspring2 = new SCPChromosomes(parent1->getProblem());

    if(debug) cout << " 2. I am in SCPCrossoverUniform::apply() " << endl;

    // Build binary representation
    tSet<int> a1 = parent1->getColumns();
    for(i=0;i<a1.getSize();i++) if( a1.getElementAtSafe(i,j)) cpr1.set(j);

    a1 = parent2->getColumns();
    for(i=0;i<a1.getSize();i++) if( a1.getElementAtSafe(i,j)) cpr2.set(j);

    for(i=0;i<numOfCols;i++) 
    {
       r = 0.01 * (rand()%101);

       if( r>=0.5 )
       {
         if( cpr1.test(i) ) coff1.set(i);
         if( cpr2.test(i) ) coff2.set(i);
       }
       else
	 {
            if( cpr1.test(i) ) coff2.set(i);
            if( cpr2.test(i) ) coff1.set(i);
	 }

    }// end of for

    for(i=0;i<numOfCols;i++) if( coff1.test(i) ) offspring1->addColumn(i);

    for(i=0;i<numOfCols;i++) if( coff2.test(i) ) offspring2->addColumn(i);

    repair(offspring1);
    repair(offspring2);

    /*if(debug) 
    { 
       cout << " parent1   : " << *parent1 << endl;
       cout << " parent2   : " << *parent2 << endl;
       cout << " offspring1 : " << *offspring1 << endl;
       cout << " offspring2 : " << *offspring2 << endl;

    }// end of if
    */

    offspring1->reduce();
    offspring2->reduce();

    /*if(debug) 
    {
       ofstream ou1("/tmp/uniform.out",ios::app);
       ou1  << " parent1   : " << *parent1 << endl;
       ou1  << " parent2   : " << *parent2 << endl;
       ou1  << " offspring1 : " << *offspring1 << endl;
       ou1  << " offspring2 : " << *offspring2 << endl;
       ou1.close();
    }// end of if
    */

    ChromosomesBag ret(2);
    ret.addChromosomes(offspring1);
    ret.addChromosomes(offspring2);

    return ret;

}// end of apply()

/*************************************************************************/
/*************************************************************************/

ChromosomesBag SCPCrossoverUniform::apply(Chromosomes* par1, Chromosomes * par2, float percent)
{   
    return apply(par1,par2);

}// end of apply()

/*************************************************************************/
/*************************************************************************/
void SCPCrossoverUniform::repair(SetCoverSolution* asol) { asol->restoreRandom(); }

